from django.db import models

from .config import DB_COMPATIBLE
from .fieldlogger import log_fields as _log_fields
from .fieldlogger import set_primary_keys


class FieldLoggerManager(models.Manager):
    def bulk_create(
        self, objs, log_fields: bool = True, run_callbacks: bool = True, **kwargs
    ):
        ignore_conflicts = kwargs.get("ignore_conflicts", False)
        if ignore_conflicts or (
            not DB_COMPATIBLE and isinstance(self.model._meta.pk, models.AutoField)
        ):
            set_primary_keys(objs, self.model)

        res = super().bulk_create(objs, **kwargs)

        if log_fields:
            _log_fields(self.model, objs, run_callbacks=run_callbacks)

        return res

    def bulk_update(
        self,
        objs,
        fields,
        log_fields: bool = True,
        run_callbacks: bool = True,
        **kwargs,
    ):
        if log_fields:
            pre_instances = self.in_bulk([obj.pk for obj in objs])

        res = super().bulk_update(objs, fields, **kwargs)

        if log_fields:
            for obj in objs:
                obj._fieldlogger_pre_instance = pre_instances.get(obj.id)

            _log_fields(
                self.model, objs, update_fields=fields, run_callbacks=run_callbacks
            )

            for obj in objs:
                del obj._fieldlogger_pre_instance

        return res



from functools import reduce
from typing import Optional

from django.db.models import Model


def getrmodel(cls: Model, rfield: str) -> Optional[Model]:
    rfield = rfield.replace(".", "__").split("__")

    def _getrmodel(c, attr):
        return getattr(c, attr).field.related_model if hasattr(c, attr) else None

    return reduce(_getrmodel, rfield, cls)


def hasrmodel(cls: Model, rfield: str) -> bool:
    rfield = rfield.replace(".", "__").split("__")

    def _hasrmodel(c, attr):
        return hasattr(c, attr) and getattr(c, attr).field.related_model

    return bool(reduce(_hasrmodel, rfield, cls))



from typing import Dict, FrozenSet, Iterable

from django.core.exceptions import FieldDoesNotExist
from django.db import transaction
from django.db.models import Max
from django.db.models.fields import DecimalField, Field

from .config import DB_COMPATIBLE, get_config
from .models import Callback, FieldLog, LoggableModel


def set_primary_keys(objs, model_class):
    with transaction.atomic():
        max_id = model_class.objects.aggregate(max_id=Max("pk"))["max_id"] or 0
        for i, obj in enumerate(objs):
            obj.id = max_id + i + 1


def _log_fields(
    instances: Iterable[LoggableModel], logging_fields: FrozenSet[Field] = None
) -> Dict[Field, FieldLog]:
    logs = {}
    field_logs_to_create = []

    for instance in instances:
        pre_instance = getattr(instance, "_fieldlogger_pre_instance", None)

        for field in logging_fields:
            try:
                new_value = getattr(instance, field.name)
                if isinstance(field, DecimalField):
                    new_value = FieldLog.from_db_field(field, new_value)

                old_value = getattr(pre_instance, field.name) if pre_instance else None

            except (FieldDoesNotExist, AttributeError):
                continue

            if new_value == old_value:
                continue

            field_log = FieldLog(
                app_label=instance._meta.app_label,
                model_name=instance._meta.model_name,
                instance_id=instance.pk,
                field=field.name,
                old_value=old_value,
                new_value=new_value,
                created=pre_instance is None,
            )

            field_logs_to_create.append(field_log)
            logs.setdefault(instance.pk, {})[field] = field_log

    if field_logs_to_create:
        if not DB_COMPATIBLE:
            set_primary_keys(field_logs_to_create, FieldLog)
        FieldLog.objects.bulk_create(field_logs_to_create)

    return logs


def _run_callbacks(
    instances: Iterable[LoggableModel],
    callbacks: Iterable[Callback],
    logs: Iterable[Dict[Field, FieldLog]],
    logging_fields: FrozenSet[Field] = None,
    fail_silently: bool = False,
):
    for instance in instances:
        for callback in callbacks:
            try:
                callback(instance, logging_fields, logs.get(instance.pk, {}))
            except Exception as e:
                if not fail_silently:
                    raise e


def log_fields(
    sender: LoggableModel,
    instances: Iterable[LoggableModel],
    update_fields: FrozenSet[Field] = None,
    run_callbacks: bool = True,
) -> Dict[Field, FieldLog]:
    LOGGING_CONFIG = get_config()

    if sender not in LOGGING_CONFIG:
        return {}

    logging_config = LOGGING_CONFIG[sender]
    logging_fields = logging_config.get("logging_fields", frozenset())
    if update_fields:
        logging_fields = frozenset(
            field for field in logging_fields if field.name in update_fields
        )

    callbacks = logging_config.get("callbacks", [])
    fail_silently = logging_config.get("fail_silently", False)

    logs = _log_fields(instances, logging_fields)

    if run_callbacks:
        _run_callbacks(instances, callbacks, logs, logging_fields, fail_silently)

    return logs



from django.db.models.signals import post_save, pre_save

from .config import get_config
from .fieldlogger import log_fields


def pre_save_log_fields(sender, instance, *args, **kwargs):
    if instance.pk:
        instance._fieldlogger_pre_instance = sender.objects.filter(
            pk=instance.pk
        ).first()


def post_save_log_fields(sender, instance, created, *args, **kwargs):
    update_fields = kwargs["update_fields"] or frozenset()

    # Log fields
    log_fields(sender, [instance], update_fields)

    # Clean up
    if hasattr(instance, "_fieldlogger_pre_instance"):
        del instance._fieldlogger_pre_instance


for model_class in get_config():
    pre_save.connect(pre_save_log_fields, model_class)
    post_save.connect(post_save_log_fields, model_class)



from copy import deepcopy

import pytest
from django.conf import settings

from .helpers import (
    CREATE_FORM,
    UPDATE_FORM,
    bulk_check_logs,
    bulk_set_attributes,
    check_logs,
    reload_modules,
    set_attributes,
    set_config,
)
from .testapp.models import TestModel, TestModelRelated

ORIGINAL_SETTINGS = deepcopy(settings.FIELD_LOGGER_SETTINGS)


@pytest.fixture
def test_instance(expected_count):
    UPDATE_FORM["test_related_field"] = TestModelRelated.objects.create()

    instance = TestModel.objects.create(
        **CREATE_FORM, test_related_field=TestModelRelated.objects.create()
    )

    check_logs(instance, len(CREATE_FORM) + 1, created=True)
    yield instance
    check_logs(instance, expected_count)


@pytest.mark.django_db(transaction=True)
class TestCase1:
    @pytest.mark.parametrize("expected_count", [0])
    def test_log_on_create(self, test_instance, expected_count):
        pass

    @pytest.mark.parametrize("update_fields", [False, True])
    @pytest.mark.parametrize("expected_count", [len(UPDATE_FORM)])
    def test_log_on_save(self, test_instance, update_fields, expected_count):
        set_attributes(test_instance, UPDATE_FORM, update_fields)

    @pytest.mark.parametrize("update_fields", [False, True])
    @pytest.mark.parametrize("expected_count", [len(UPDATE_FORM)])
    def test_log_on_save_twice(self, test_instance, update_fields, expected_count):
        set_attributes(test_instance, UPDATE_FORM, update_fields)
        set_attributes(test_instance, UPDATE_FORM, update_fields)


@pytest.fixture
def restore_settings():
    yield
    settings.FIELD_LOGGER_SETTINGS = deepcopy(ORIGINAL_SETTINGS)
    reload_modules()


@pytest.mark.django_db(transaction=True)
@pytest.mark.usefixtures("restore_settings")
@pytest.mark.parametrize("scope", ["global", "testapp", "testmodel"])
class TestCase2:
    def test_logging_disabled(self, scope):
        set_config({"logging_enabled": False}, scope)
        test_instance = TestModel.objects.create(**CREATE_FORM)
        check_logs(test_instance, expected_count=0, created=True)

    def test_fail_silently(self, scope):
        set_config({"fail_silently": False, "callbacks": [lambda *args: 1 / 0]}, scope)
        with pytest.raises(ZeroDivisionError):
            TestModel.objects.create(**CREATE_FORM)


@pytest.fixture
def test_instances(expected_count, log_fields, run_callbacks, ignore_conflicts):
    related_instance = TestModelRelated.objects.create()

    UPDATE_FORM["test_related_field"] = TestModelRelated.objects.create()

    instances = TestModel.objects.bulk_create(
        [
            TestModel(test_related_field=related_instance, **CREATE_FORM)
            for _ in range(5)
        ],
        log_fields=log_fields,
        run_callbacks=run_callbacks,
        ignore_conflicts=ignore_conflicts,
    )

    bulk_check_logs(
        instances,
        len(CREATE_FORM) + 1 if log_fields else 0,
        run_callbacks,
        created=True,
    )
    yield instances
    bulk_check_logs(instances, expected_count if log_fields else 0, run_callbacks)


@pytest.mark.django_db(transaction=True)
@pytest.mark.parametrize("log_fields", [True, False])
@pytest.mark.parametrize("run_callbacks", [True, False])
@pytest.mark.parametrize("ignore_conflicts", [False, True])
class TestCase3:
    @pytest.mark.parametrize("expected_count", [0])
    def test_log_on_bulk_create(self, test_instances, log_fields, run_callbacks):
        pass

    @pytest.mark.parametrize("expected_count", [len(UPDATE_FORM)])
    def test_log_on_bulk_update(self, test_instances, log_fields, run_callbacks):
        bulk_set_attributes(test_instances, UPDATE_FORM, save=False)

        TestModel.objects.bulk_update(
            test_instances,
            UPDATE_FORM.keys(),
            log_fields=log_fields,
            run_callbacks=run_callbacks,
        )

    @pytest.mark.parametrize("expected_count", [len(UPDATE_FORM)])
    def test_log_on_bulk_update_twice(self, test_instances, log_fields, run_callbacks):
        bulk_set_attributes(test_instances, UPDATE_FORM, save=False)

        TestModel.objects.bulk_update(
            test_instances,
            UPDATE_FORM.keys(),
            log_fields=log_fields,
            run_callbacks=run_callbacks,
        )

        bulk_set_attributes(test_instances, UPDATE_FORM, save=False)

        TestModel.objects.bulk_update(
            test_instances,
            UPDATE_FORM.keys(),
            log_fields=log_fields,
            run_callbacks=run_callbacks,
        )



import pytest

from fieldlogger.utils import getrmodel, hasrmodel

from .helpers import CREATE_FORM
from .testapp.models import TestModel, TestModelRelated, TestModelRelated2


@pytest.fixture
def test_instance():
    related_instance3 = TestModel.objects.create(**CREATE_FORM)
    related_instance2 = TestModelRelated2.objects.create(
        **CREATE_FORM, test_related_field3=related_instance3
    )
    related_instance = TestModelRelated.objects.create(
        **CREATE_FORM, test_related_field2=related_instance2
    )
    instance = TestModel.objects.create(
        **CREATE_FORM, test_related_field=related_instance
    )
    instance.test_attribute = "test"
    return instance


@pytest.mark.django_db
@pytest.mark.parametrize(
    "field", [field for field in TestModel._meta.fields if field.name != "id"]
)
class TestUtilsOnDirectFields:
    def test_getrmodel(self, field):
        assert getrmodel(TestModel, field.name) == field.related_model

    def test_hasrmodel(self, field):
        assert hasrmodel(TestModel, field.name) == field.is_relation


@pytest.mark.django_db
@pytest.mark.parametrize("sep", [".", "__"])
@pytest.mark.parametrize(
    "related_field",
    [
        (["test_related_field", "test_char_field"], None),
        (["test_related_field", "test_related_field2"], TestModelRelated2),
        (
            ["test_related_field", "test_related_field2", "test_char_field"],
            None,
        ),
        (
            ["test_related_field", "test_related_field2", "test_related_field3"],
            TestModel,
        ),
        (
            [
                "test_related_field",
                "test_related_field2",
                "test_related_field3",
                "test_char_field",
            ],
            None,
        ),
    ],
)
class TestUtilsOnRelatedFields:
    def test_getrmodel(self, sep, related_field):
        rattr, expected_cls = related_field
        assert getrmodel(TestModel, sep.join(rattr)) == expected_cls
        assert (
            expected_cls
            == getattr(
                getrmodel(TestModel, sep.join(rattr[:-1])),
                rattr[-1],
            ).field.related_model
        )

    def test_hasrmodel(self, sep, related_field):
        rattr, expected_cls = related_field
        assert hasrmodel(TestModel, sep.join(rattr)) == bool(expected_cls)


@pytest.mark.django_db
@pytest.mark.parametrize("sep", [".", "__"])
@pytest.mark.parametrize(
    "related_field",
    [
        [""],
        ["non_existent_field"],
        ["test_related_field", "non_existent_field"],
        ["test_related_field", "test_related_field2", "non_existent_field"],
    ],
)
class TestUtilsOnNonExistentFields:
    def test_getrmodel(self, sep, related_field):
        assert getrmodel(TestModel, sep.join(related_field)) is None

    def test_hasrmodel(self, sep, related_field):
        assert not hasrmodel(TestModel, sep.join(related_field))



from datetime import date, datetime, time, timedelta
from decimal import Decimal
from json import JSONDecoder, JSONEncoder
from uuid import UUID

from django.conf import settings
from django.core.files import File
from django.db import models
from django.utils.module_loading import import_string

SETTINGS = getattr(settings, "FIELD_LOGGER_SETTINGS", {})


class Encoder(JSONEncoder):
    def default(self, obj):
        if isinstance(obj, (date, datetime, time)):
            return obj.isoformat()
        elif isinstance(obj, timedelta):
            return str(obj.total_seconds())
        elif isinstance(obj, Decimal):
            return float(obj)
        elif isinstance(obj, (bytes, bytearray)):
            return obj.decode()
        elif isinstance(obj, memoryview):
            return obj.tobytes().decode()
        elif isinstance(obj, File):
            return obj.name
        elif isinstance(obj, UUID):
            return str(obj)
        elif isinstance(obj, models.Model):
            return obj.pk
        elif isinstance(obj, models.QuerySet):
            return list(obj.values_list("pk", flat=True))

        return super().default(obj)


class Decoder(JSONDecoder):
    pass


ENCODER = SETTINGS.get("ENCODER")
ENCODER = import_string(ENCODER) if ENCODER else Encoder

DECODER = SETTINGS.get("DECODER")
DECODER = import_string(DECODER) if DECODER else Decoder



# Generated by Django 5.1.4 on 2024-12-26 22:39

import fieldlogger.encoding
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="FieldLog",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("app_label", models.CharField(editable=False, max_length=100)),
                ("model_name", models.CharField(editable=False, max_length=100)),
                ("instance_id", models.CharField(editable=False, max_length=255)),
                (
                    "field",
                    models.CharField(
                        editable=False, max_length=100, verbose_name="field name"
                    ),
                ),
                ("timestamp", models.DateTimeField(auto_now=True)),
                (
                    "old_value",
                    models.JSONField(
                        blank=True,
                        decoder=fieldlogger.encoding.Decoder,
                        editable=False,
                        encoder=fieldlogger.encoding.Encoder,
                        null=True,
                    ),
                ),
                (
                    "new_value",
                    models.JSONField(
                        blank=True,
                        decoder=fieldlogger.encoding.Decoder,
                        editable=False,
                        encoder=fieldlogger.encoding.Encoder,
                        null=True,
                    ),
                ),
                (
                    "extra_data",
                    models.JSONField(
                        decoder=fieldlogger.encoding.Decoder,
                        default=dict,
                        encoder=fieldlogger.encoding.Encoder,
                    ),
                ),
                ("created", models.BooleanField(default=False, editable=False)),
            ],
        ),
    ]



from functools import cached_property
from typing import Callable, Dict, FrozenSet, NewType

from django.apps import apps
from django.core.exceptions import FieldDoesNotExist
from django.db import models
from django.utils.translation import gettext_lazy as _

from .encoding import DECODER, ENCODER
from .utils import getrmodel


class FieldLog(models.Model):
    app_label = models.CharField(max_length=100, editable=False)
    model_name = models.CharField(max_length=100, editable=False)
    instance_id = models.CharField(max_length=255, editable=False)
    field = models.CharField(_("field name"), max_length=100, editable=False)
    timestamp = models.DateTimeField(auto_now=True, editable=False)
    old_value = models.JSONField(
        encoder=ENCODER, decoder=DECODER, blank=True, null=True, editable=False
    )
    new_value = models.JSONField(
        encoder=ENCODER, decoder=DECODER, blank=True, null=True, editable=False
    )
    extra_data = models.JSONField(encoder=ENCODER, decoder=DECODER, default=dict)
    created = models.BooleanField(default=False, editable=False)

    def __str__(self):
        return (
            f"({self.app_label}__{self.model_name}__{self.field}, created={self.created})\
        {self.old_value} -> {self.new_value}"
        )

    @staticmethod
    def from_db_field(field_class, value):
        if field_class.__class__ is models.BinaryField:
            value = bytes(value, "utf-8") if isinstance(value, str) else value
        elif field_class.__class__ is models.DecimalField:
            value = round(value, field_class.decimal_places) if value else value
        elif field_class.__class__ is models.ForeignKey:
            return field_class.related_model.objects.get(pk=value) if value else None

        return field_class.to_python(value)

    @classmethod
    def from_db(cls, db, field_names, values):
        model_class = apps.get_model(values[1], values[2])
        fieldpath, _, field = values[4].rpartition("__")
        model_class = getrmodel(model_class, fieldpath) or model_class

        try:
            field_class = model_class._meta.get_field(field)
        except FieldDoesNotExist:
            field_class = None

        values[3] = model_class._meta.pk.to_python(values[3])

        if field_class is not None:
            values[6] = (
                cls.from_db_field(field_class, values[6]) if not values[9] else None
            )
            values[7] = cls.from_db_field(field_class, values[7])
        else:
            values[6] = None if not values[9] else values[6]
            values[7] = values[7]

        return super().from_db(db, field_names, values)

    @cached_property
    def model(self):
        return apps.get_model(self.app_label, self.model_name)

    @cached_property
    def instance(self):
        return self.model.objects.get(pk=self.instance_id)

    @cached_property
    def previous_log(self):
        return (
            self.__class__.objects.filter(
                app_label=self.app_label,
                model_name=self.model_name,
                instance_id=self.instance_id,
                field=self.field,
            )
            .exclude(pk=self.pk)
            .order_by("pk")
            .last()
        )


LoggableModel = NewType("LoggableModel", models.Model)
Callback = Callable[[LoggableModel, FrozenSet[str], Dict[str, FieldLog]], None]



from functools import reduce
from typing import Any, Dict, FrozenSet, List, Union

from django import VERSION
from django.apps import apps
from django.conf import settings
from django.db import connection
from django.db.models.fields import Field
from django.utils.module_loading import import_string

from .models import Callback, LoggableModel

DB_ENGINE = connection.vendor
if DB_ENGINE == "sqlite":
    DB_VERSION = connection.Database.sqlite_version_info
elif DB_ENGINE == "mysql":
    DB_VERSION = connection.Database.mysql_version
else:
    DB_VERSION = None

DB_COMPATIBLE = VERSION >= (4, 0) and (
    DB_ENGINE == "postgresql"
    or (DB_ENGINE == "mysql" and DB_VERSION >= (10, 5))
    or (DB_ENGINE == "sqlite" and DB_VERSION >= (3, 35))
)


class LoggingConfig:
    __config = {}

    def __init__(self):
        self.__settings = getattr(settings, "FIELD_LOGGER_SETTINGS", {})

    def __cfg_reduce(self, op, key, *configs, default=None):
        return reduce(
            op,
            [config.get(key, default) for config in configs],
            self.__settings.get(key.upper(), default),
        )

    def __logging_enabled(self, *configs: Dict[str, bool]) -> bool:
        return self.__cfg_reduce(
            lambda a, b: a and b, "logging_enabled", *configs, default=True
        )

    def __logging_fields(
        self, model_class: LoggableModel, model_config: Dict[str, Any]
    ) -> FrozenSet[Field]:
        fields = model_config.get("fields", [])
        exclude_fields = set(model_config.get("exclude_fields", []))
        model_fields = model_class._meta.get_fields()

        if fields == "__all__":
            return frozenset(
                field for field in model_fields if field.name not in exclude_fields
            )

        return frozenset(
            field
            for field in model_fields
            if field.name in set(fields) - exclude_fields
        )

    def __callbacks(
        self, *configs: Dict[str, List[Union[str, Callback]]]
    ) -> List[Callback]:
        callbacks = self.__cfg_reduce(
            lambda a, b: a + b, "callbacks", *configs, default=[]
        )

        callbacks = [
            import_string(callback) if isinstance(callback, str) else callback
            for callback in callbacks
        ]

        return callbacks

    def __fail_silently(self, *configs: Dict[str, bool]) -> bool:
        return self.__cfg_reduce(
            lambda a, b: a and b, "fail_silently", *configs, default=True
        )

    def __set_config(self):
        for app, app_config in self.__settings.get("LOGGING_APPS", {}).items():
            if not app_config or not self.__logging_enabled(app_config):
                continue

            for model, model_config in app_config.get("models", {}).items():
                if not model_config or not self.__logging_enabled(
                    app_config, model_config
                ):
                    continue

                try:
                    model_class = apps.get_model(app, model)
                except LookupError:
                    continue

                self.__config[model_class] = {
                    "logging_fields": self.__logging_fields(model_class, model_config),
                    "callbacks": self.__callbacks(app_config, model_config),
                    "fail_silently": self.__fail_silently(app_config, model_config),
                }

    def get_config(self):
        if not self.__config:
            self.__set_config()

        return self.__config

    def __iter__(self):
        return iter(self.__config)

    def __getitem__(self, model_class: LoggableModel):
        return self.__config.get(model_class, {})


get_config = LoggingConfig().get_config



from django.apps import AppConfig


class FieldloggerConfig(AppConfig):
    default_auto_field = "django.db.models.BigAutoField"
    name = "fieldlogger"

    def ready(self):
        import fieldlogger.signals  # noqa: F401



from functools import cached_property

from django.db import models

from .models import FieldLog


class FieldLoggerMixin(models.Model):
    @cached_property
    def fieldlog_set(self):
        return FieldLog.objects.filter(
            instance_id=self.pk,
            model_name=self._meta.model_name,
            app_label=self._meta.app_label,
        )

    class Meta:
        abstract = True
